import Data.Map as M hiding (map, foldr)
import Data.Maybe
import Control.Monad

table = M.fromList [("a",2), ("b",3), ("c",4)]

lookupA :: Maybe Int
lookupA = M.lookup "a" table -- => Just 2

lookupZ :: Maybe Int
lookupZ = M.lookup "z" table -- => Nothing

-- a は取れないかもしれないのでこれを考慮する
aPlus1 :: Maybe Int
aPlus1 = let a = M.lookup "a" table
         in if isJust a
            then Just (1 + (fromJust a))
            else Nothing

aPlus1' :: Maybe Int
aPlus1' = case M.lookup "a" table of
  Nothing -> Nothing
  Just a -> Just (a + 1)

-- Just や Nothing が消えた!(抽象化!!)
aPlus1'' :: Maybe Int
aPlus1'' = fmap (\a -> a + 1) (M.lookup "a" table)

-- fmap f mA = case mA of
--    Nothing -> Nothing
--    Just a  -> Just (f a)

-- join mmA = case mmA of
--   Nothing       -> Nothing
--   Just Nothing  -> Nothing
--   Just (Just a) -> Just a

mA >>= f = case mA of
  Nothing = Nothing
  Just a = Just (f a)

aPlusB :: Maybe Int
aPlusB = case M.lookup "a" table of
  Nothing -> Nothing
  Just a  -> case M.lookup "b" table of
    Nothing -> Nothing
    Just b -> Just (a + b) -- Just

-- 抽象化? Just (Just 5) つらい…
aPlusB' :: Maybe (Maybe Int)
aPlusB' = fmap (\a ->
                 fmap (\b -> -- aを使えるところで書かないといけない
                        a + b
                      ) (M.lookup "b" table)
                 ) (M.lookup "a" table) -- Just (Just 5)

-- join の導入
aPlusB'' :: Maybe Int
aPlusB'' = join . fmap (\a ->
                 fmap (\b ->
                        a + b
                      ) (M.lookup "b" table)
                 ) $ (M.lookup "a" table)

aPlusBPlusC :: Maybe Int
aPlusBPlusC =
  case M.lookup "a" table of
    Nothing -> Nothing
    Just a  -> case M.lookup "b" table of
      Nothing -> Nothing
      Just b -> case M.lookup "c" table of
        Nothing -> Nothing
        Just c -> Just $ a + b + c

aPlusBPlusC' :: Maybe (Maybe (Maybe Int))
aPlusBPlusC' = fmap (\a ->
                 fmap (\b ->
                   fmap (\c ->
                     a + b + c
                   ) (M.lookup "c" table)
                 ) (M.lookup "b" table)
               ) (M.lookup "a" table)


aPlusBPlusC_' :: Maybe Int
aPlusBPlusC_' = join . fmap (\a ->
                 join . fmap (\b ->
                    fmap (\c ->
                        a + b + c
                      )   (M.lookup "c" table)
                    ) $ (M.lookup "b" table)
                 ) $ (M.lookup "a" table)

aPlusB''' :: Maybe Int
aPlusB''' = join . fmap (\a ->
              join . fmap (\b ->
                Just $ a + b
              ) $ (M.lookup "b" table)
            ) $ (M.lookup "a" table)

aPlusBPlusC'' :: Maybe Int
aPlusBPlusC'' = join . fmap (\a ->
                  join . fmap (\b ->
                    join . fmap (\c ->
                        Just $ a + b + c
                      ) $ (M.lookup "c" table)
                    ) $ (M.lookup "b" table)
                  ) $ (M.lookup "a" table)


aPlusBPlusC''' :: Maybe Int
aPlusBPlusC''' = (>>= (\a ->
                   (>>= (\b ->
                     (>>= (\c ->
                        return $ a + b + c
                     )) (M.lookup "c" table)
                   )) (M.lookup "b" table)
                 )) (M.lookup "a" table)


aPlusBPlusC'''' :: Maybe Int
aPlusBPlusC'''' =
  M.lookup "a" table >>= \a ->
    M.lookup "b" table >>= \b ->
      M.lookup "c" table >>= \c ->
        return $ a + b + c

aPlusBPlusC''''' :: Maybe Int
aPlusBPlusC''''' = do
  a <- M.lookup "a" table
  b <- M.lookup "b" table
  c <- M.lookup "c" table
  return $ a + b + c

aPlusBPlusC_ mA mB mC = do
 a <- mA
 b <- mB
 c <- mC
 return $ a + b + c


aPlusBPlusC_bind mA mB mC = do
 mA >>= \a ->
   mB >>= \b ->
     mC >>= \c ->
     return $ a + b + c

-- aPlusBPlusC_ (M.lookup "a" table) (M.lookup "c" table) (M.lookup "c" table)

-- mA >>= f = case mA of
--   Nothing = Nothing
--   Just a = f a

aPlusBPlusC__ mA mB mC =
  case mA of
    Nothing -> Nothing
    Just a -> case mB of
      Nothing -> Nothing
      Just b -> case mC of
        Nothing -> Nothing
        Just c -> Just (a + b + c)

(>>=) :: [a] -> (a -> [b]) -> [b]
mA >>= f =
  case mA of
    [] -> []
    x:xs -> go (f x) (xs >>= f)
      where
        go [] acc = acc
        go (y:ys) acc = y : (go ys acc)


apbpc =
  case mA of
    [] -> []
    xa:xas -> go ((\a ->
                    case mB of
                      [] -> []
                    xb:xbs -> go ((\b ->
                      case mC of
                        [] -> []
                      x:xs -> go ((\c -> a + b + c) x) (xs >>= (\c -> a + b + c))
                      where
                        go [] acc = acc
                        go (y:ys) acc = y : (go ys acc)
                                  ) xb) (xbs >>= (\b ->
                                                   case mC of
                                                     [] -> []
                                                   x:xs -> go ((\c -> a + b + c) x) (xs >>= (\c -> a + b + c))
                                                   where
                                                     go [] acc = acc
                                                     go (y:ys) acc = y : (go ys acc)
                                ))
                     where
                       gob [] accb = accb
                       gob (yb:ybs) accb = y : (go ybs accb)
                  ) xa) (xas >>= (\a ->
                                    ))
      where
        go [] acca = acca
        go (ya:yas) acc = y : (go yas acc
  case mC of
    [] -> []
    x:xs -> go ((\c -> a + b + c) x) (xs >>= (\c -> a + b + c))
      where
        go [] acc = acc
        go (y:ys) acc = y : (go ys acc)

omap = flip fmap

listAPlusBPlusC :: [] ([] ([] Int)) -- [[[Int]]]
listAPlusBPlusC = [2,3] `omap` \a ->
                    [3,4] `omap` \b ->
                      [4,5] `omap` \c ->
                         a + b + c

{- inRuby
[2,3].map do |a|
  [3,4].map do |b|
    [4].map do |c|
      a + b + c
  end
end
-}
-- [[[9,10],[10,11]],[[10,11],[11,12]]]

-- [Int]にしつつ、抽象化
listAPlusBPlusC' :: [Int]
listAPlusBPlusC' = [2,3] >>= \a ->
                     [3,4] >>= \b ->
                       [4,5] >>= \c ->
                         return $ a + b + c

-- doでかく
listAPlusBPlusC'' :: [Int]
listAPlusBPlusC'' = do
  a <- [2,3]
  b <- [3,4]
  c <- [4,5]
  return $ a + b + c


-- aPlusBPlusC_ [2,3] [3,4] [4,5]           -- [9,10,10,11,10,11,11,12]
-- aPlusBPlusC_ (Just 2) (Just 3) (Just 4)  -- Just 9
-- aPlusBPlusC_ (Just 2) Nothing (Just 4)   -- Nothing
-- aPlusBPlusC_ [2,3] [] [4,5]              -- []
